from requests import Session
from robobrowser import RoboBrowser
import boto3
import xml.etree.ElementTree as ET
from os.path import expanduser
import configparser
import getpass
import base64
from . import configuration, utils


class AbstractADFS(object):
    def __init__(self, logger, config):
        self.logger = logger
        self.config = config
        self.configure = configuration.Configure(logger)
        if not self.config.has_section('aws_accounts'):
            self.config.add_section('aws_accounts')
        self.aws_accounts = config._sections['aws_accounts']

    def get_username_password(self):
        print ('Please enter your federation credentials')
        self.username = self.configure.input_and_set(self.config,'msadfs','username','Username')
        self.password = getpass.getpass()
        print ('')
        return self.username, self.password

    def delete_username_password(self):
        self.username = '##############################################'
        self.password = '##############################################'
        del self.username
        del self.password

    def init_browser(self):
        # Using RoboBrowser to get the SAML token
        self.browser = RoboBrowser(parser='lxml')
        self.browser.open(self.config.get('provider','idpentryurl'))
        return self.browser

    def handle_saml(self):
        self.logger.debug("handling saml response and finding out aws roles")
        browser = self.browser
        # Decode the response and extract the SAML assertion
        assertion = ''

        # Look for the SAMLResponse attribute of the input tag (determined by
        # analyzing the debug print lines above)
        for inputtag in browser.find_all('input'):
            if(inputtag.get('name') == 'SAMLResponse'):
                #print(inputtag.get('value'))
                assertion = inputtag.get('value')

        # Better error handling is required for production use.
        if (assertion == ''):
            #TODO: Insert valid error checking/handling
            print ('Response did not contain a valid SAML assertion')
            sys.exit(0)

        # Debug only
        #print(base64.b64decode(assertion))

        # Parse the returned assertion and extract the authorized roles
        awsroles = []
        root = ET.fromstring(base64.b64decode(assertion))
        for saml2attribute in root.iter('{urn:oasis:names:tc:SAML:2.0:assertion}Attribute'):
            if (saml2attribute.get('Name') == 'https://aws.amazon.com/SAML/Attributes/Role'):
                for saml2attributevalue in saml2attribute.iter('{urn:oasis:names:tc:SAML:2.0:assertion}AttributeValue'):
                    awsroles.append(saml2attributevalue.text)

        # Note the format of the attribute value should be role_arn,principal_arn
        # but lots of blogs list it as principal_arn,role_arn so let's reverse
        # them if needed
        for awsrole in awsroles:
            chunks = awsrole.split(',')
            if'saml-provider' in chunks[0]:
                newawsrole = chunks[1] + ',' + chunks[0]
                index = awsroles.index(awsrole)
                awsroles.insert(index, newawsrole)
                awsroles.remove(awsrole)

        # If I have more than one role, ask the user which one they want,
        # otherwise just proceed
        print ("")
        if len(awsroles) > 1:
            i = 0
            print ("Please choose the role you would like to assume:")
            for awsrole in awsroles:
                role = awsrole.split(',')[0]
                roleName = role.split(':')[5]
                accountId = role.split(':')[4]
                if accountId in self.aws_accounts:
                    print ('[{:2d}]: {:10s} {:30s}:{:10s}'.format(i,accountId, self.aws_accounts.get(accountId),roleName))
                else:
                    print ('[', i, ']: ', role)
                i += 1

            selectedroleindex = self.configure.input_and_set(config=self.config,section='msadfs',option='selectedroleindex',label='Selection')

            # Basic sanity check of input
            if int(selectedroleindex) > (len(awsroles) - 1):
                print ('You selected an invalid role index, please try again')
                sys.exit(0)

            role_arn = awsroles[int(selectedroleindex)].split(',')[0]
            principal_arn = awsroles[int(selectedroleindex)].split(',')[1]
        else:
            role_arn = awsroles[0].split(',')[0]
            principal_arn = awsroles[0].split(',')[1]

        self.configure.store_config(self.config)

        # Use the assertion to get an AWS STS token using Assume Role with SAML
        stsResponse = boto3.client('sts').assume_role_with_saml(
            RoleArn=role_arn,
            PrincipalArn=principal_arn,
            SAMLAssertion=assertion,
            DurationSeconds=3600 # is currently the maximum
        )

        # Write the AWS STS token into the AWS credential file
        filename = self.config.get('aws','credentials_file')

        # Read in the existing config file
        awsconfig = configparser.RawConfigParser()
        awsconfig.read(filename)

        # We are using the specified profile in the config file to write the credentials
        profile = self.config.get('provider','profile_name')

        if not awsconfig.has_section(profile):
            awsconfig.add_section(profile)
        else:
            # we should backup the profile ones
            if not awsconfig.has_section(profile+'.backup'):
                awsconfig.add_section(profile+'.backup')
                items = awsconfig.items(profile)
                for item in items:
                    awsconfig.set(profile+'.backup', item[0], item[1])

        awsconfig.set(profile, 'output', self.config.get('aws', 'outputformat'))
        awsconfig.set(profile, 'region', self.config.get('aws', 'region'))
        awsconfig.set(profile, 'aws_access_key_id', stsResponse['Credentials']['AccessKeyId'])
        awsconfig.set(profile, 'aws_secret_access_key', stsResponse['Credentials']['SecretAccessKey'])
        awsconfig.set(profile, 'aws_session_token', stsResponse['Credentials']['SessionToken'])

        # Write the updated config file
        with open(filename, 'w+') as configfile:
            awsconfig.write(configfile)

        # Give the user some basic info as to what has just happened
        print ('\n\n----------------------------------------------------------------')
        print ('Your new access key pair has been stored in the AWS configuration file {0} under the {1} profile.'.format(filename, profile))
        print ('Note that it will expire at {0}.'.format(stsResponse['Credentials']['Expiration']))
        print ('After this time, you may safely rerun this script to refresh your access key pair.')
        print ('To use this credential, call the AWS CLI with the --profile option (e.g. aws --profile saml ec2 describe-instances) unless you defined the default profile in configuration')
        print ('----------------------------------------------------------------\n\n')



from . import abstract_adfs

class MicrosoftADFS:

    def __init__(self, logger, config):
        logger.info("initialize MicrosoftADFS")
        self.logger = logger
        self.config = config
        self.adfs = abstract_adfs.AbstractADFS(logger,config)

        username, password = self.adfs.get_username_password()

        browser = self.adfs.init_browser()
        if (browser.find(id='loginForm')):
            form = browser.get_form(id='loginForm')
            form["UserName"] = username
            form["Password"] = password
        else:
            print ('Could not find the required forms. Maybe different provider')
            sys.exit(1)

        # Submitting the form
        browser.submit_form(form)

        self.adfs.delete_username_password()
        self.adfs.handle_saml()



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import sys
import argparse

def init():
    """ process command line arguments """
    parser = argparse.ArgumentParser(description="Request temporary AWS account credentials "
                                                 "for a federated account",
                                     usage="%(prog)s [options]",
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog="""
NOTE: If you have several aws accounts where you have access to, you can map them in the config file (~/.aws/adfs_auth.ini) to a suitable name.
EXAMPLE:
...
[aws_accounts]
123456789 = Production
234567890 = Staging
345678901 = Development
...

""")

    parser.add_argument("-V", "--version",
                        dest="show_version", action="store_true",
                        help="Show the version number and exit")

    parser.add_argument("-C", "--configure",
                        dest="configure", action="store_true",
                        help="Configures the AWS ADFS Auth application")

    parser.add_argument("-v", "--verbose",
                        dest="verbosity", action="count",
                        help="Output debug messages, increase messages with -v -v")
    return parser

def show_version():
    """ show version and exit """
    print("aws_adfs_auth version:{}".format(sys.modules["aws_adfs_auth"].VERSION))
    sys.exit(0)

def error(message):
    """ display an error related to command line arguments and quit """
    print(message + "\nhint: try -h for help")
    sys.exit(1)

def check_args():
    """ checking the command line arguments """
    parser = init()
    parsed_options = parser.parse_args()

    if parsed_options.show_version:
        show_version()


    return parsed_options



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import os
import configparser
from os.path import expanduser


class Configure(object):
    home = expanduser("~")
    aws_folder = home + '/.aws'
    aws_adfs_auth_config_file = aws_folder + '/adfs_auth.ini'
    adfs_poviders = ['Microsoft']

    def __init__(self, logger):
        self.logger = logger

    def open_config(self):
        self.logger.debug('Opening config from file %s' % self.aws_adfs_auth_config_file)
        config = configparser.RawConfigParser()
        config.read(self.aws_adfs_auth_config_file)
        return config

    def check_config(self):
        """ function to check if the configuration is set """
        self.logger.debug('checking config folders and files')
        if not os.path.exists(self.aws_folder):
            os.makedirs(self.aws_folder)
            self.logger.info("aws folder doesn't exist. App is not configured")
            return False
        if not os.path.isfile(self.aws_adfs_auth_config_file):
            self.logger.info("aws config file doesn't exist. App is not configured")
            return False
        config = self.open_config()
        if (not config.has_section('provider') or not
               config.has_option('provider', 'name') or not
               config.has_option('provider', 'idpentryurl')):
            self.logger.info("sanity check failed on config file. App is not configured")
            return False
        return True

    def store_config(self, config):
        """ function to store the configuration """
        self.logger.debug("storing config file to %s" % self.aws_adfs_auth_config_file)
        with open(self.aws_adfs_auth_config_file, 'w+') as adfs_auth_config_file:
            config.write(adfs_auth_config_file)

    def setup_provider(self, config):
        """ function to setup the provider """
        print ("Please choose which ADFS provider you want to use (currently only Microsoft is supported)")
        if len(self.adfs_poviders) > 1:
            i = 0
            for adfs_provider in self.adfs_poviders:
                print ('[{:2d}]: {:30s}'.format(i, adfs_provider))
                i += 1
            selected_provider_index = int(input('Selection: '))
        else:
            selected_provider_index = 0
        if not config.has_section('provider'):
            config.add_section('provider')
        config.set('provider', 'name', self.adfs_poviders[selected_provider_index])

    def setup_idpentryurl(self, config):
        """ function to setup the idp entry url """
        self.logger.debug("setting up idp url")
        print ("Please enter the ADFS provider enty URL for your configuration")
        print ("Example: https://<provider>/adfs/ls/IdpInitiatedSignOn.aspx?loginToRp=urn:amazon:webservices")
        self.input_and_set(config=config, section='provider', option='idpentryurl', label='URL')

    def setup_profile_name(self, config):
        """ function to setup the profile name to use for storing the credentials """
        self.logger.debug("setting up the profile name")
        print ("Please enter the AWS profile you want to use for storing your credentials")
        print ("If you use the default profile you don't need to specify the profile when executing AWS commands.")
        print ("Make a backup of your default profile if you still need it.")
        self.input_and_set(config=config, section='provider', option='profile_name', label='AWS Profile')

    def input_and_set(self, config, section, option, label):
        """ function to ask for input and if the value is
        already set it will use the same value on pressing enter """
        if not config.has_section(section):
            config.add_section(section)
        if config.has_option(section, option):
            last_value = config.get(section, option)
            value = input('{:s} [{:s}]: '.format(label, last_value)) or last_value
        else:
            value = input('{:s}: '.format(label))
        config.set(section, option, value)
        return value

    def set_value(self, config, section, option, value):
        """ function to set a value within the section """
        if not config.has_section(section):
            config.add_section(section)
        if not config.has_option(section, option):
            config.set(section, option, value)

    def setup_variables(self, config):
        """ function to setup additional variables """
        print ("Please enter the AWS region")
        self.input_and_set(config=config, section='aws', option='region', label='Region')
        self.input_and_set(config=config, section='aws', option='outputformat', label='Output format')

    def setup_defaults(self, config):
        """ function to setup defaults """
        self.set_value(config, 'aws', 'credentials_file', self.home + '/.aws/credentials')
        self.set_value(config, 'aws', 'sslverification', True)
        self.set_value(config, 'info', 'version', '0.3.0')
        self.set_value(config, 'provider', 'profile_name', 'saml')

    def setup(self):
        """ function to setup the configuration of adfs auth """
        config = self.open_config()
        self.setup_provider(config)
        self.setup_idpentryurl(config)
        self.setup_variables(config)
        self.setup_defaults(config)
        self.setup_profile_name(config)
        self.store_config(config)
        return config

    def migrate_020_030(self, config):
        self.set_value(config, 'info', 'version', '0.3.0')
        self.set_value(config, 'provider', 'profile_name', 'saml')
        self.store_config(config)

    def migrate(self, config):
        """ function to migrate the configuration to the next version """
        self.logger.info('migrating configuration file to the latest version if necessary')
        # checking if config has a version number, if not we start to migrate from 0.2.0
        if not config.has_section('info'):
            config.add_section('info')
        if not config.has_option('info', 'version'):
            version = '0.2.0'
        else:
            version = config.get('info', 'version')
        if version == '0.3.0':
            # all good, we are on the latest version
            self.logger.info('migration of configuration not necessary.')
        elif version == '0.2.0':
            self.logger.info('migrating configuration from 0.2.0 to 0.3.0')
            self.migrate_020_030(config)
        else:
            raise Exception('configuration file corrupted, please re-configure application')



""" adfs_auth module definition """

#from . import configuration
#from . import cli
#from . import main

VERSION = "0.3.0"
BUILD_TIME = "2018-01-26 01:25:00"

__title__ = "aws_adfs_auth"
__version__ = VERSION
__summary__ = "aws_adfs_auth command line utility"
__license__ = "APACHE2"
__uri__ = "https://github.com/jschwellach/aws-adfs-auth"
__author__ = "Janos Schwellach"
__email__ = "jschwellach@gmail.com"



from setuptools import setup
import sys
import os
import aws_adfs_auth

install_requires = [
    'markdown',
    'requests',
    'robobrowser',
    'boto3',
    'lxml',
    'pytest-runner'
]

tests_requires = [
    'pytest'
]


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = aws_adfs_auth.__title__,
    version = aws_adfs_auth.__version__,
    description = aws_adfs_auth.__summary__,
    long_description=read('README.md'),
    license = aws_adfs_auth.__license__,
    url = aws_adfs_auth.__uri__,
    author = aws_adfs_auth.__author__,
    author_email = aws_adfs_auth.__email__,
    packages=['aws_adfs_auth'],
    install_requires=install_requires,
    tests_require=tests_requires,
    extras_requires={},
    data_files= [("", ["LICENSE"])],
    entry_points={'console_scripts': ['aws_adfs_auth = aws_adfs_auth.main:main']},
    keywords = "",
    test_suite = 'tests.aws_adfs_auth_test_suite',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
)
